import os
import fc_oogmsh

def get_geodirs(dim,d):
  s=get_pathname(dim,d)
  assert s is not None,"Unable to find geo directory for dim=%d and d=%d"%(dim,d)
  fullname=os.path.dirname(os.path.abspath(__file__))
  loc_dir=fullname+os.sep+'geodir'+os.sep+s
  return [loc_dir,fc_oogmsh.sys.get_geodirs(dim,d)]
  
def get_pathname(dim,d):
  if dim==2 and d==2:
    return '2d'
  if dim==3 and d==2:
    return '3ds'
  if dim==3 and d==3:
    return '3d'
  return None


import matplotlib.pyplot as plt
import numpy as np
#from mpl_toolkits.mplot3d import Axes3D
from scipy import linalg,sparse
from scipy.sparse.linalg import spsolve
import itertools
from math import factorial
#import matplotlib
#from matplotlib.patches import Polygon,Patch
#from mpl_toolkits.mplot3d import Axes3D
#import mpl_toolkits.mplot3d as a3
#from mpl_toolkits.mplot3d.art3d import Poly3DCollection,Line3DCollection
from fc_tools.colors import selectColors
#from fc_tools.graphics import isMayavi

def comb(n,k):
  import itertools
  return np.array([x for x  in itertools.combinations(np.arange(n),k)],dtype=int)

def issiMeshElt(sTh):
  return isinstance(sTh,siMeshElt)

class siMeshElt:
  def __init__(self):
     self.d=0
     self.dim=0
     self.nq=0
     self.nme=0
     self.q=None
     self.me=None
     self.dglobal=0
     self.toGlobal=None
     self.nqGlobal=0
     self.toParent=None
     self.nqParent=0
     self.nqParents=[] # nqParents[i]  i=0: parent, i=1:parent of parent, ...
     self.toParents=[]
     self.label=0
     self.Tag=''
     self.color=[]
     self.vols=None
     self.gradBaCo=None
     self.geolab =[]
     self.partlab=[]
     self.bbox=[]
     self.h=0
     
  def __init__(self,dim,d,label,q,me,toGlobal,nqGlobal,geolab,partlab):
     self.dim=dim
     self.d=d
     self.label=label
     self.q=q
     self.nq=q.shape[0]
     self.me=me
     self.nme=me.shape[1]
     self.dglobal=0
     self.toGlobal=toGlobal
     self.nqGlobal=nqGlobal
     self.toParent=toGlobal 
     self.nqParent=nqGlobal
     self.nqParents=[nqGlobal] # nqParents[i]  i=0: parent, i=1:parent of parent, ...
     self.toParents=[toGlobal]
     self.geolab=geolab
     self.partlab=partlab
     self.color=[]
     self.Tag=''
     self.ComputeVolVec()
     self.GetMaxLengthEdges()
     self.GradBaCo()
     #self.vols=ComputeVolVec(self.d,self.q,self.me)
     #self.gradBaCo=GradBaCo(self.d,self.q,self.me)
     #self.h=GetMaxLengthEdges(self.q,self.me)
     self.bbox=np.zeros((2*dim,)) # bounding box
     for i in range(dim):
       self.bbox[2*i]=min(q[:,i])
       self.bbox[2*i+1]=max(q[:,i])
     
  def __repr__(self):
    strret = ' %s object \n'%self.__class__.__name__ 
    strret += '      d : %d\n'%self.d 
    strret += '    dim : %d\n'%self.dim 
    strret += '  label : %d\n'%self.label
    strret += '     nq : %d\n'%self.nq
    strret += '    nme : %d\n'%self.nme
    strret += 'nqGlobal: %d\n'%self.nqGlobal
    strret += '      q : size=%s\n'%self.q.shape.__repr__()
    strret += '     me : size=%s\n'%self.me.shape.__repr__()
    strret += 'toGlobal: size=%s\n'%self.toGlobal.shape.__repr__()
    return strret  
  
  def feval(self,fun,**kwargs):
    Global=kwargs.get('global', False)
    dtype=kwargs.get('dtype', float)
    assert np.isscalar(fun) or isinstance(fun, type(lambda: None)) or isinstance(fun,np.lib.function_base.vectorize), "1st argument must be a scalar, a def function or a lambda function. Given a %s"%str(fun)
    nargin=get_func_nargin(fun)
    if nargin is not None: # fun is a 'def' function or a 'lambda' function
      assert nargin==1 or nargin==self.dim, "1st argument is a function function and must have 1 or %d input arguments"%nargin
      if nargin==self.dim:
        fun=np.vectorize(fun)
        S='fun(self.q[:,0]'
        for i in range(1,self.dim):
          S=S+',self.q[:,'+str(i)+']'
        S=S+')'
        V=eval(S)
      else:
        V=fun(self.q.T)
      fun=V # With constant functions, V is a scalar!
    if np.isscalar(fun):
      if Global:
        V=fun*np.ones((self.nqGlobal,),dtype=dtype)
      else:
        V=fun*np.ones((self.nq,),dtype=dtype)
    return V
  
  def eval(self,f,**kwargs):
    """ Eval f on siMeshElt object and return a numpy array with shape (self.nq,)
      corresponding on the evaluate of f on each mesh vertices (if f is not a 
      list and not None)
      
      f could be:
        - a scalar
        - a def function
        - a lambda function
        - a numpy.vectorize function
        - a numpy array 
        - a list where each element could be on of the previous type
    """
    if f is None:
      return None
    dtype=kwargs.get('dtype', float)
    Num=kwargs.get('Num', 1)
    if np.isscalar(f):
      return f*np.ones((self.nq,),dtype=dtype)
    if isinstance(f, type(lambda: None)) or isinstance(f,np.lib.function_base.vectorize):
      V=self.feval(f)
      return self.feval(f)
    if isinstance(f,np.ndarray) and (f.shape[0]==self.nq) :
      return f
    if isinstance(f,list): # Vector Field case
      n=len(f)
      Fh=np.zeros((n*self.nq,));I=np.arange(self.nq)
      VFInd=getVFindices(Num,n,self.nq)
      for i in range(n):
        Fh[VFInd(I,i)]=self.eval(f[i])
      return Fh
    else:
      assert False
    

  def ComputeVolVec(self):
    # AJHanson_Geometry_for_N-Dimensional, (5) page 156 Graphics Gems IV edited by Paul S. Heckbert, 
    # published by Academic Press Limited
    # Introduction to the geometry of N Dimensions by D. M. Y. Sommerville (1929)
    if self.d==0:
      self.vols=1
      return
    d=self.d;q=self.q;me=self.me;nme=self.nme;
    X=np.zeros((d,nme,self.dim))
    #print(X.shape)
    for i in range(d):
      X[i]=q[me[i+1]]-q[me[0]]
    V=np.zeros((d,d,nme))
    for i in range(d):
      V[i,i]=(X[i]*X[i]).sum(axis=1)
      for j in range(i+1,d):
        V[i,j]=V[j,i]=(X[i]*X[j]).sum(axis=1)
    self.vols=np.array([np.sqrt(abs(linalg.det(V[::,::,k])))/factorial(d) for k in range(nme)])
 
  def GetMaxLengthEdges(self):
    ne=self.d+1;q=self.q;me=self.me
    h=0.
    for i in range(ne):
      for j in range(i+1,ne):
        h=max(h,np.sum((q[me[i]]-q[me[j]])**2,axis=1).max())
    self.h=np.sqrt(h)
    
  def GradBaCo(self):
    d=self.d;dim=self.dim
    nme=self.nme;q=self.q;me=self.me
    A=np.zeros((d,dim,nme))
    I=np.zeros((d,dim,nme),dtype=int)
    J=np.zeros((d,dim,nme),dtype=int)
    L=np.arange(nme)
    for i in range(d):
      A[i]=(q[me[i+1]]-q[me[0]]).T
      I[i]=np.tile(L*d+i,(dim,1))
      for j in range(dim):
        J[i,j]=L*dim+j
    N=d*dim*nme
    spA=sparse.csc_matrix((np.reshape(A,N),(np.reshape(I,N),np.reshape(J,N))),shape=(d*nme,dim*nme))
    spH=spA*spA.T
    Grad=np.hstack([-np.ones((d,1)),np.eye(d)])
    b=np.tile(Grad,(nme,1))
    #G=spsolve(spH,b)
    GradBaCo=spA.T*spsolve(spH,b)
    self.GradBaCo=GradBaCo.reshape((nme,dim,d+1)).swapaxes(1,2)
  
  def NormalFaces(self):
    d=self.d
    #(IndLocFaces,IndOpositePt)=getIndLocFaces(d)
    Normal=np.zeros((d+1,self.dim,self.nme))
    for i in range(d+1):
      #A=-self.GradBaCo[:,IndOpositePt[i],:]
      A=-self.GradBaCo[:,i,:]
      Normal[i,:,:]=A.T
      N2=np.sqrt(np.sum(A**2,axis=1))
      for j in range(self.dim):
        Normal[i,j,:]=Normal[i,j,:]/N2
    return Normal
      
  def barycenters(self):
    Ba=np.zeros((self.dim,self.nme))
    for i in range(self.d+1):
      Ba+=self.q[self.me[i]].T
    Ba/=(self.d+1)
    return Ba
  
  def move(self,U):
    self.q=move_mesh(self,U)
  
def move_mesh(sTh,U):
  Q=sTh.q
  if U is None:
    return Q
  assert(len(U)==sTh.dim)
  if len(U[0])==sTh.nqGlobal:
    for i in range(sTh.dim):
      Q[:,i]+=U[i][sTh.toGlobal]
  else:
    assert( len(U[0]) == sTh.nq )
    for i in range(sTh.dim):
      Q[:,i]+=U[i]
  return Q    
  
def getIndLocFaces(d):
  IndLocFaces=comb(d+1,d)
  IndOpositePt=np.zeros(d+1,dtype=int)
  I=np.arange(d+1)
  for i in range(d+1):
    IndOpositePt[i]=np.setdiff1d(I,IndLocFaces[i])
  return  IndLocFaces,IndOpositePt

def mesh_label(sTh):
  assert issiMeshElt(sTh), "First argument must be a siMeshElt object"
  return 'd=%d, lab=%d'%(sTh.d,sTh.label)

def getLocalValue(sTh,u):    
  if len(u)==sTh.nqGlobal:
    U=u[sTh.toGlobal]
  elif len(u)==sTh.nqParent:
    U=u[sTh.toParent]
  else:
    assert( len(u) == sTh.nq )
    U=u
  return U.ravel()

def getLocalVector(sTh,V): # V : list of n vectors
  assert isinstance(V,list)
  n=len(V)
  Vl=n*[None]
  for i in range(n):
    Vl[i]=getLocalValue(sTh,V[i])
  return Vl

def get_func_nargin(f):
  #assert  isinstance(f, type(lambda: None)) or isinstance(f,np.lib.function_base.vectorize), "1st argument must be a def function or a lambda function. Given a %s"%str(fun)
  if isinstance(f, type(lambda: None)): # fun is a 'def' function or a 'lambda' function
    return f.__code__.co_argcount
  if isinstance(f,np.lib.function_base.vectorize):
    return f.pyfunc.__code__.co_argcount
  return None


from fc_oogmsh import gmsh
from fc_simesh.siMesh import siMesh,HyperCube
import numpy as np


def feval2D01():
  meshfile=gmsh.buildmesh2d('condenser11',50)
  Th=siMesh(meshfile)  

  g1=lambda x,y: np.cos(x)*np.sin(y)
  g2=lambda X: np.cos(X[0])*np.sin(X[1])

  def g3(x,y):
    return np.cos(x)*np.sin(y)

  def g4(X):
    return np.cos(X[0])*np.sin(X[1])

  z1=Th.feval(g1)
  z2=Th.feval(g2)
  z3=Th.feval(g3)
  z4=Th.feval(g4)

  print('max(abs(z2-z1))=%e'%max(abs(z2-z1)))
  print('max(abs(z3-z1))=%e'%max(abs(z3-z1)))
  print('max(abs(z4-z1))=%e'%max(abs(z4-z1)))
  
  
def feval3D01():  
  meshfile=gmsh.buildmesh3d('cylinderkey',15)
  Th=siMesh(meshfile)
  g1=lambda x,y,z: 3*x**2-y**3+z**2+x*y
  g2=lambda X: 3*X[0]**2-X[1]**3+X[2]**2+X[0]*X[1]
  
  def g3(x,y,z):
    return 3*x**2-y**3+z**2+x*y
  
  def g4(X):
    return 3*X[0]**2-X[1]**3+X[2]**2+X[0]*X[1]
  
  z1=Th.feval(g1)
  z2=Th.feval(g2)
  z3=Th.feval(g3)
  z4=Th.feval(g4)

  print('max(abs(z2-z1))=%e'%max(abs(z2-z1)))
  print('max(abs(z3-z1))=%e'%max(abs(z3-z1)))
  print('max(abs(z4-z1))=%e'%max(abs(z4-z1)))
  
def feval3Ds01():  
  meshfile=gmsh.buildmesh3ds('demisphere5',30)
  Th=siMesh(meshfile)
  g1=lambda x,y,z: 3*x**2-y**3+z**2+x*y
  g2=lambda X: 3*X[0]**2-X[1]**3+X[2]**2+X[0]*X[1]
  
  def g3(x,y,z):
    return 3*x**2-y**3+z**2+x*y
  
  def g4(X):
    return 3*X[0]**2-X[1]**3+X[2]**2+X[0]*X[1]
  
  z1=Th.feval(g1)
  z2=Th.feval(g2)
  z3=Th.feval(g3)
  z4=Th.feval(g4)

  print('max(abs(z2-z1))=%e'%max(abs(z2-z1)))
  print('max(abs(z3-z1))=%e'%max(abs(z3-z1)))
  print('max(abs(z4-z1))=%e'%max(abs(z4-z1)))
  
def feval2D02():
  meshfile=gmsh.buildmesh2d('condenser11',50)
  Th=siMesh(meshfile)  
  f=lambda x,y: np.cos(2*x)*np.sin(3*y)

  def g(x,y,cx,cy): 
    return np.cos(cx*x)*np.sin(cy*y)

  g1=lambda x,y: g(x,y,2,3)
  g2=lambda X: g(X[0],X[1],2,3)

  z=Th.feval(f)
  z1=Th.feval(g1)
  z2=Th.feval(g2)

  print('max(abs(z1-z))=%e'%max(abs(z1-z)))
  print('max(abs(z2-z))=%e'%max(abs(z2-z)))
  
  
def feval3D02():  
  meshfile=gmsh.buildmesh3d('cylinderkey',15)
  Th=siMesh(meshfile)
  f=lambda x,y,z: 3*x**2-2*y**3+4*z**2-x*y
  
  def g(x,y,z,a,b,c,d):
    return a*x**2+b*y**3+c*z**2+d*x*y
  
  g1=lambda x,y,z: g(x,y,z,3,-2,4,-1)
  g2=lambda X: g(X[0],X[1],X[2],3,-2,4,-1)

  z=Th.feval(f)
  z1=Th.feval(g1)
  z2=Th.feval(g2)

  print('max(abs(z1-z))=%e'%max(abs(z1-z)))
  print('max(abs(z2-z))=%e'%max(abs(z2-z)))
  
def feval3Ds02():  
  meshfile=gmsh.buildmesh3ds('demisphere5',30)
  Th=siMesh(meshfile)
  f=lambda x,y,z: 3*x**2-2*y**3+4*z**2-x*y
  
  def g(x,y,z,a,b,c,d):
    return a*x**2+b*y**3+c*z**2+d*x*y
  
  g1=lambda x,y,z: g(x,y,z,3,-2,4,-1)
  g2=lambda X: g(X[0],X[1],X[2],3,-2,4,-1)

  z=Th.feval(f)
  z1=Th.feval(g1)
  z2=Th.feval(g2)

  print('max(abs(z1-z))=%e'%max(abs(z1-z)))
  print('max(abs(z2-z))=%e'%max(abs(z2-z)))  


def feval2D03():
  meshfile=gmsh.buildmesh2d('condenser11',50)
  Th=siMesh(meshfile)  
  # f : R^2 -> R^3
  f=[lambda x,y: np.cos(2*x)*np.sin(3*y),
    lambda x,y: np.cos(3*x)*np.sin(4*y),
    lambda x,y: np.cos(4*x)*np.sin(5*y)]
  
  g=[lambda X: np.cos(2*X[0])*np.sin(3*X[1]),
    lambda x,y: np.cos(3*x)*np.sin(4*y),
    lambda x,y: np.cos(4*X[0])*np.sin(5*X[1])]
  
  V1=Th.feval(f)
  V2=Th.feval(f)
  print('V1.shape='+str(V1.shape)+' V2.shape='+str(V2.shape))
  print('max(abs(V1-V2))=%e'%np.max(np.abs(V1-V2)))
  
def feval3D03():  
  meshfile=gmsh.buildmesh3d('cylinderkey',15)
  Th=siMesh(meshfile)
    # f : R^3 -> R^2
  f=[lambda x,y,z: np.cos(2*x)*np.sin(3*y)-z,
    lambda x,y,z: np.cos(3*x)*np.sin(4*y)+z]
  
  g=[lambda X: np.cos(2*X[0])*np.sin(3*X[1])-X[2],
    lambda x,y,z: np.cos(3*x)*np.sin(4*y)+z]
  
  V1=Th.feval(f)
  V2=Th.feval(f)
  print('V1.shape='+str(V1.shape)+' V2.shape='+str(V2.shape))
  print('max(abs(V1-V2))=%e'%np.max(np.abs(V1-V2)))
  
  
def feval3Ds03():  
  meshfile=gmsh.buildmesh3ds('demisphere5',30)
  Th=siMesh(meshfile)
    # f : R^3 -> R^2
  f=[lambda x,y,z: np.cos(2*x)*np.sin(3*y)-z,
    lambda x,y,z: np.cos(3*x)*np.sin(4*y)+z]
  
  g=[lambda X: np.cos(2*X[0])*np.sin(3*X[1])-X[2],
    lambda x,y,z: np.cos(3*x)*np.sin(4*y)+z]
  
  V1=Th.feval(f)
  V2=Th.feval(f)
  print('V1.shape='+str(V1.shape)+' V2.shape='+str(V2.shape))
  print('max(abs(V1-V2))=%e'%np.max(np.abs(V1-V2)))
  
def find2D01():
  meshfile=gmsh.buildmesh2d('condenser11',50)
  Th=siMesh(meshfile)
  print('*** 1: print mesh')
  print(Th)
  print('*** 2: print labels of the 2-simplex elementary meshes')
  idx=Th.find(2)
  print('d=2, labels='+str(Th.sThlab[idx]))
  print('*** 3: print labels of the 1-simplex elementary meshes')
  idx=Th.find(1)
  print('d=1, labels='+str(Th.sThlab[idx]))
  d=2;lab=6
  print('*** 4: print %d-simplex elementary meshes with label %d'%(d,lab))
  idx=Th.find(d,labels=lab)
  print('Th.sThlab[%d]=%s'%(idx,Th.sThlab[idx]))
  print('Th.sTh[%d] is the '%idx + str(Th.sTh[idx]))
  d=1;lab=101
  print('*** 5: print %d-simplex elementary meshes with label %d'%(d,lab))
  idx=Th.find(d,labels=lab)
  print('Th.sThlab[%d]=%s'%(idx,Th.sThlab[idx]))
  print('Th.sTh[%d] is the '%idx + str(Th.sTh[idx]))
  
def find3D01():  
  meshfile=gmsh.buildmesh3d('cylinder3dom',15)
  Th=siMesh(meshfile)
  print('*** 1: print mesh')
  print(Th)
  d=3
  print('*** 2: print labels of the %d-simplex elementary meshes'%d)
  idx=Th.find(d)
  print('d=%d, labels='%d +str(Th.sThlab[idx]))
  d=2
  print('*** 3: print labels of the %d-simplex elementary meshes'%d)
  idx=Th.find(d)
  print('d=%d, labels='%d +str(Th.sThlab[idx]))
  d=1
  print('*** 4: print labels of the %d-simplex elementary meshes'%d)
  idx=Th.find(d)
  print('d=%d, labels='%d +str(Th.sThlab[idx]))
  d=3;lab=10
  print('*** 5: print %d-simplex elementary meshes with label %d'%(d,lab))
  idx=Th.find(d,labels=lab)
  print('Th.sThlab[%d]=%s'%(idx,Th.sThlab[idx]))
  print('Th.sTh[%d] is the '%idx + str(Th.sTh[idx]))
  d=2;lab=112
  print('*** 6: print %d-simplex elementary meshes with label %d'%(d,lab))
  idx=Th.find(d,labels=lab)
  print('Th.sThlab[%d]=%s'%(idx,Th.sThlab[idx]))
  print('Th.sTh[%d] is the '%idx + str(Th.sTh[idx]))
  d=1;lab=120
  print('*** 7: print %d-simplex elementary meshes with label %d'%(d,lab))
  idx=Th.find(d,labels=lab)
  print('Th.sThlab[%d]=%s'%(idx,Th.sThlab[idx]))
  print('Th.sTh[%d] is the '%idx + str(Th.sTh[idx]))
  
def find3Ds01():  
  meshfile=gmsh.buildmesh3ds('demisphere5',30)
  Th=siMesh(meshfile)
  print('*** 1: print mesh')
  print(Th)
  d=2
  print('*** 2: print labels of the %d-simplex elementary meshes'%d)
  idx=Th.find(d)
  print('d=%d, labels='%d +str(Th.sThlab[idx]))
  d=1
  print('*** 3: print labels of the %d-simplex elementary meshes'%d)
  idx=Th.find(d)
  print('d=%d, labels='%d +str(Th.sThlab[idx]))
  d=0
  print('*** 4: print labels of the %d-simplex elementary meshes'%d)
  idx=Th.find(d)
  print('d=%d, labels='%d +str(Th.sThlab[idx]))
  d=2;lab=10
  print('*** 5: print %d-simplex elementary meshes with label %d'%(d,lab))
  idx=Th.find(d,labels=lab)
  print('Th.sThlab[%d]=%s'%(idx,Th.sThlab[idx]))
  print('Th.sTh[%d] is the '%idx + str(Th.sTh[idx]))
  d=1;lab=5
  print('*** 6: print %d-simplex elementary meshes with label %d'%(d,lab))
  idx=Th.find(d,labels=lab)
  print('Th.sThlab[%d]=%s'%(idx,Th.sThlab[idx]))
  print('Th.sTh[%d] is the '%idx + str(Th.sTh[idx]))
  d=0;lab=13
  print('*** 7: print %d-simplex elementary meshes with label %d'%(d,lab))
  idx=Th.find(d,labels=lab)
  print('Th.sThlab[%d]=%s'%(idx,Th.sThlab[idx]))
  print('Th.sTh[%d] is the '%idx + str(Th.sTh[idx]))
      




#import sys;sys.path.insert(1,'/home/cuvelier/Travail/Recherch/python/fc-hypermesh')
#import sys;sys.path.insert(1,'/home/cuvelier/Travail/Recherch/python/fc-tools')
#import sys;sys.path.insert(1,'/home/cuvelier/Travail/Recherch/python/fc-oogmsh')




import sys
import numpy as np
from math import factorial
from fc_simesh.siMeshElt import siMeshElt
from fc_oogmsh.gmsh import ooGmsh,EltTypeSimplex
from fc_tools.colors import selectColors
from fc_hypermesh.OrthMesh import OrthMesh

# Medit format Name
simplices_dict={'Vertices':0,'Edges':1,'Triangles':2,'Tetrahedra':3}

def issiMesh(Th):
  return isinstance(Th,siMesh)

class siMesh:    
  def __init__(self,gmsh_file=None,**kwargs):
     self.d=0
     self.dim=0
     self.order=0
     self.sTh=[]
     self.nsTh=0
     self.toGlobal=[]
     self.toParent=[]
     self.sThsimp=[]
     self.sThlab=[]
     self.sThcolors=[]
     self.bbox=[] # Bounding box
     self.sThgeolab=[]
     self.sThphyslab=[]
     self.sThpartlabs={}
     self.M=[]
     self.sElts=[]
     self.nq=0
     if gmsh_file==None:
       return
     try:
       fid = open(gmsh_file, "r")
     except IOError:
       print("File '%s' not found." % (gmsh_file))
       sys.exit()
     isPhysical=kwargs.get('isPhysical', True)
     dim=kwargs.get('dim', 2)
     d=kwargs.get('d', None)
     trans=kwargs.get('trans', None)
     self.readgmsh(gmsh_file,dim,d,isPhysical,trans)
     self.set_all()
          
  def __repr__(self):
    strret = ' %s object \n'%self.__class__.__name__ 
    strret += '      d : %d\n'%self.d 
    strret += '    dim : %d\n'%self.dim
    strret += '    nq  : %d\n'%self.nq
    strret += '    nme : %d\n'%self.get_nme()
    strret += '    sTh : %s of %d %s\n'%(self.sTh.__class__.__name__,len(self.sTh),self.sTh[0].__class__.__name__)
    strret += '   nsTh : %d\n'%self.nsTh
    strret += 'sThsimp : %s %s\n'%(str(self.sThsimp.shape),self.sThsimp.__class__.__name__)
    strret += '   '+np.array_str(self.sThsimp).replace('\n','\n          ')+'\n'
    strret += ' sThlab : %s %s\n'%(str(self.sThlab.shape),self.sThlab.__class__.__name__)
    strret += '   '+np.array_str(self.sThlab).replace('\n','\n          ')+'\n'
    return strret         
  
  def readgmsh(self,gmsh_file,dim,d,isPhysLabel,trans):
    self.dim=dim
    G=ooGmsh(gmsh_file)
    #Q=G.q
    if trans is not None:
      if trans.__code__.co_argcount==1:
        G.q=trans(G.q)
      elif trans.__code__.co_argcount==2:
        G.q=trans(G.q[:,0],G.q[:,1])
      elif trans.__code__.co_argcount==3:
        G.q=trans(G.q[:,0],G.q[:,1],G.q[:,2])
      else:
        print('Error with trans function')
    #print(Q-G.q)  
    if G.dim!=dim:
      print('  Mesh %s is a %d-dimensional mesh\n    Force dimension to %d\n'%(gmsh_file,G.dim,G.dim))
      self.dim=G.dim
    if d==None:
      #d=[dim,dim-1]
      d=np.arange(self.dim,-1,-1)
    #self.d=max(d)
    self.toGlobal=G.toGlobal
    self.nq=G.toGlobal.shape[0]
    self.toParent=G.toGlobal
    self.M=G.M
    self.sElts=G.sElts
    
    for i in range(len(G.sElts)):
      dd=EltTypeSimplex(G.sElts[i].type)
      if (dd != None ) and (dd in d):
        nme=G.sElts[i].me.shape[1]
        if G.sElts[i].part_lab[0]==None:
          if isPhysLabel:
            mel=G.sElts[i].phys_lab
          else:
            mel=G.sElts[i].geo_lab
        else:
          print('to do')
       
        self.AddsiMeshElts(dd,G.q,nme,G.sElts[i].me,mel)#,geolab=G.sElts[i].GeometricalLabel,partlab=G.sElts[i].part_lab)
    self.sThsimp=np.array(self.sThsimp,dtype=int)
    self.sThlab=np.array(self.sThlab,dtype=int)
    self.d=max(self.sThsimp)
                           
  def AddsiMeshElts(self,d,q,nme,me,mel,**kwargs):
    Geolab=kwargs.get('geolab', None)
    Partlab=kwargs.get('partlab', None)
    labels=np.unique(mel)
    geolab=None
    partlab=None
    k=self.nsTh;
    nq=q.shape[0]
    if d==self.d:
      self.toParent=np.unique(np.concatenate((self.toParent,me.flatten())))
      self.toGlobal=self.toParent
    for i in range(len(labels)):
       I=(mel==labels[i]).nonzero()[0]
       ME=me[:,I]  
       if Geolab!=None:
         geolab=Geolab[I]
       if Partlab!=None:
         partlab=Partlab[I]
       indQ=np.unique(ME)
       Q=q[indQ]
       lQ=np.arange(0,len(indQ))
       J=np.zeros((nq,),dtype=int)
       J[indQ]=lQ
       MELoc=J[ME]
       self.sTh.append(siMeshElt(self.dim,d,labels[i],Q,MELoc,indQ,nq,geolab,partlab))
       self.sThsimp.append(d)
       self.sThlab.append(labels[i])
       self.nsTh+=1
       
       
  def AddCopysiMeshElt(self,ooElt):
    if len(np.intersect1d(self.toGlobal,ooElt.toGlobal))==len(ooElt.toGlobal):
      self.nsTh+=1
      k=self.nsTh
      from copy import deepcopy
      self.sTh.append(deepcopy(ooElt))
      #self.sThlab=np.append(self.sThlab,ooElt.label)
      #self.sThsimp=np.append(self.sThsimp,ooElt.d)
      self.sThlab.append(ooElt.label)
      self.sThsimp.append(ooElt.d)
      
  def AddsiMeshElt(self,ooElt):
    if self.sTh != []:
      #print(np.intersect1d(self.toGlobal,ooElt.toGlobal).shape)
      if not isinstance(ooElt.toGlobal,np.ndarray):
        ooElt.toGlobal=np.array([ooElt.toGlobal],dtype=int)
      #print(ooElt.toGlobal)
      #ooElt.toGlobal=np.array(ooElt.toGlobal,dtype=int)
      #print(ooElt.toGlobal)
      I=np.intersect1d(self.toGlobal,ooElt.toGlobal)
      if I.shape[0]!=1:
        #assert( len(ooElt.toGlobal)==1 )
      #else:
        assert( I.shape==ooElt.toGlobal.shape )
      #assert(len(np.intersect1d(self.toGlobal,ooElt.toGlobal))==len(ooElt.toGlobal))
      N=self.nsTh+1
    else:
      N=1;
      self.toGlobal=ooElt.toGlobal
      self.nq=ooElt.nq
    self.nsTh=N
    self.sTh.append(ooElt)
    self.sThlab.append(ooElt.label)
    self.sThsimp.append(ooElt.d)
      
  def find(self,d,labels=None):
    #sThsimp=np.array(self.sThsimp) # A l'arrache
    #sThlab=np.array(self.sThlab)
    ind=(self.sThsimp==d).nonzero()[0]
    if ind.shape[0]==0:
      return None
    if labels is None:
      return ind
    if isinstance(labels,int) or isinstance(labels,np.int64):
      I=np.where(self.sThlab[ind]==labels)[0]
      if I.shape[0]==1:
        return ind[I[0]]
      else:
        return None
    isList=False
    if isinstance(labels,list):
      islist=True
    labels=np.array(labels)
    Ind=[]
    for i in range(len(ind)):
      if self.sThlab[ind[i]] in labels:
        Ind.append(ind[i])
    if not isList:
      Ind=np.array(Ind)
    return Ind
  
  def set_nq(self):
    idxdom=self.find(self.d)
    assert(len(idxdom)>0)
    if len(idxdom)==1:
      self.nq=self.sTh[idxdom[0]].nq
      return
    toG=[]
    for i in idxdom:
      toG.append(self.sTh[i].toGlobal)
    self.nq=np.unique(np.concatenate(toG)).shape[0]
                           
  def set_colors(self):
    self.sThcolors=np.zeros((self.nsTh,3));
    for d in range(self.d+1):
      Ilab=self.find(d)
      if Ilab is not None:
        self.sThcolors[Ilab]=selectColors(len(Ilab))
        for jlab in Ilab:
          self.sTh[jlab].color=self.sThcolors[jlab]
          
  def set_bbox(self):
    self.bbox=self.sTh[0].bbox
    for k in range(self.nsTh):
      bb=self.sTh[k].bbox
      for i in range(self.dim):
        self.bbox[2*i]=min(self.bbox[2*i],bb[2*i])
        self.bbox[2*i+1]=max(self.bbox[2*i+1],bb[2*i+1])
   
  def set_dglobal(self):
    for k in range(self.nsTh):
      self.sTh[k].dglobal=self.d
      
  def set_all(self):
     #self.sThlab=np.array(self.sThlab)
     #self.sThsimp=np.array(self.sThsimp)
     self.set_nq()
     self.set_colors()
     self.set_bbox()
     self.set_dglobal()
      
  def get_nme(self):
    Ilab=self.find(self.d)
    nme=0
    for jlab in Ilab:
      nme+=self.sTh[jlab].nme
    return nme
  
  def get_h(self):
    Ilab=self.find(self.d)
    h=0
    for jlab in Ilab:
      h=max(self.sTh[jlab].h,h)
    return h
  
  def move(self,U):
    if isinstance(U,list):
      assert len(U)==self.dim
      V=np.array(U)
    elif isinstance(U,np.ndarray):
      V=U
    else:
      assert False
    assert (V.shape[0]==self.dim) and (V.shape[1]==self.nq)
    for i in range(self.nsTh):
      self.sTh[i].move(V)
  
  
  def feval(self,fun,**kwargs):
    """ Eval a def function or a lambda function on siMesh object
    """
    d=kwargs.get('d', self.d)
    if isinstance(fun,list):
      U=[]
      for f in fun:
        U.append(self.feval(f,**kwargs))
      return np.array(U)
    labels=kwargs.get('labels', self.sThlab[self.find(d)])
    U=np.zeros((self.nq,))
    for l in labels:
      k=self.find(d,labels=l)
      if k is not None:
        U[self.sTh[k].toGlobal]=self.sTh[k].feval(fun)
    return U
  
  def eval(self,f,**kwargs):
    """ Eval f on siMesh object and return a numpy array with shape (self.nq,)
      corresponding on the evaluate of f on each mesh vertices (if f is not a 
      list and not None)
      
      f could be:
        - a scalar
        - a def function
        - a lambda function
        - a numpy.vectorize function
        - a numpy array 
        - a list where each element could be on of the previous type
    """
    if f is None:
      return None
    dtype=kwargs.get('dtype', float)
    Num=kwargs.get('Num', 1)
    if np.isscalar(f):
      return f*np.ones((self.nq,),dtype=dtype)
    if isinstance(f, type(lambda: None)) or isinstance(f,np.lib.function_base.vectorize):
      V=self.feval(f)
      return self.feval(f)
    if isinstance(f,np.ndarray) and (f.shape[0]==self.nq) :
      return f
    if isinstance(f,list): # Vector Field case
      n=len(f)
      Fh=np.zeros((n*self.nq,));I=np.arange(self.nq)
      VFInd=getVFindices(Num,n,self.nq)
      for i in range(n):
        Fh[VFInd(I,i)]=self.eval(f[i])
      return Fh
    else:
      assert False
  
  def barycenters(self,**kwargs):
    d=kwargs.get('d', self.d)
    labels=kwargs.get('labels', self.sThlab[self.find(d)])
    idxlab=self.find(d,labels)
    Ba=np.zeros((self.dim,len(idxlab)))
    j=0
    for i in idxlab:
      Ba[:,j]=np.mean(self.sTh[i].barycenters(),axis=1)
      j+=1
    return Ba
  
  def merge_vertices(self,**kwargs):
    d=kwargs.pop('d', self.d)
    labels=kwargs.pop('labels', self.sThlab[self.find(d)])
    toGlobal=[]
    Q=np.zeros((self.nq,self.dim))
    for l in labels:
      k=self.find(d,labels=[l])
      if len(k)==1:
        toGlobal=np.hstack((toGlobal,self.sTh[k[0]].toGlobal))
        Q[self.sTh[k[0]].toGlobal]=self.sTh[k[0]].q 
    toGlobal=np.int32(np.unique(toGlobal))
    Q=Q[toGlobal].T
    return Q,toGlobal
  
  def get_mesh(self,**kwargs):
    """
      
    """
    d=kwargs.pop('d', self.d)
    labels=kwargs.pop('labels', self.sThlab[self.find(d)])
    idxlab=self.find(d,labels=labels)
    if not isinstance(idxlab,list) and not isinstance(idxlab,np.ndarray):
      idxlab=[idxlab]
    toGlobal=[]
    ME=np.zeros((d+1,0),dtype=int)
    Q=np.zeros((self.nq,self.dim))
    for k in idxlab:
      toGlobal=np.hstack((toGlobal,self.sTh[k].toGlobal))
      Q[self.sTh[k].toGlobal]=self.sTh[k].q
      ME=np.hstack((ME,self.sTh[k].toGlobal[self.sTh[k].me]))
      
    toGlobal=np.int32(np.unique(toGlobal))
    Q=Q[toGlobal]
    N=len(toGlobal)
    idx=np.zeros((self.nq,),dtype=int)
    idx[toGlobal]=np.arange(N,dtype=int)
    ME=idx[ME]
    return Q,ME,toGlobal

     
  #from fc_simesh_matplotlib.siMesh import plotmesh,plot,plotGradBaCo,plotiso
  #if isMayavi(): 
    ##from fc_simesh.toolsmesh.vtk import vtkPlot,vtkPlotVal
    #from fc_simesh_mayavi.siMesh import vtk_plotmesh, vtk_slicemesh, vtk_plot, vtk_iso_surface, vtk_slice, vtk_sliceiso, vtk_plotiso, vtk_quiver, vtk_streamline
      
def LabelBaseName(Th,d):
  if (Th.d==Th.dim):
    if (d==Th.d):
      return r"\Omega"
    if (d+1==Th.d):
      return r"\Gamma"
    if (d+2==Th.d):
      return r"\partial\Gamma"
    if (d+3==Th.d):
      return r"\partial^2\Gamma"
  if (Th.d+1==Th.dim):
    if (d==Th.d):
      return r"\Omega"
    if (d+1==Th.d):
      return r"\Gamma"
    if (d+2==Th.d):
      return r"\partial\Gamma"

def HyperCube(d,N,**kwargs):
  mapping=kwargs.get('mapping',None)
  m_min=kwargs.get('m_min',0)
  assert(m_min in range(d+1))
  Oh=OrthMesh(d,N,mapping=mapping,m_min=m_min)
  ooTh=siMesh()
  ooTh.d=ooTh.dim=d
  Th=Oh.Mesh
  ooElt=siMeshElt(d,d,1,Th.q.T,Th.me,np.arange(Th.nq),Th.nq,None,None)
  ooTh.AddsiMeshElt(ooElt)

  for i in range(len(Oh.Faces)):
    for j in range(len(Oh.Faces[i])):
      sTh=Oh.Faces[i][j]
      ooElt=siMeshElt(d,sTh.m,j+1,sTh.q.T,sTh.me,sTh.toGlobal,Th.nq,None,None)
      ooTh.AddsiMeshElt(ooElt)
      
  ooTh.sThsimp=np.array(ooTh.sThsimp)
  ooTh.sThlab=np.array(ooTh.sThlab)
  ooTh.set_all()
  return ooTh

#def vtk_SaveAllFigsAsFiles(filename,**kwargs):
  #tag=kwargs.get('tag', False )
  #figext=kwargs.get('format', 'png' )
  #savedir=kwargs.get('dir', '.' )
  #figs=kwargs.get('figs',None) # list of figures number
  #verbose=kwargs.get('verbose',False)
  #scale=kwargs.get('scale', 1.0 )
  #import sys,os,mayavi
  #from mayavi import mlab
  #if tag:
    #Softname='Python'
    #V=sys.version.split(' ')[0]
    #Release=V.replace('.','')
    #Tag=Softname+Release
    ## FullTag=Tag+'_Mayavi'+mayavi.__version__.replace('.','') 
    
  #if not os.path.exists(savedir):
    #os.makedirs(savedir)
  
  #for i in range(len(figs)):
    #nfig=figs[i]
    #if tag:
      #File=savedir+os.path.sep+filename+'_fig'+str(nfig)+'_'+Tag+'.'+figext
    #else:
      #File=savedir+os.path.sep+filename+'_fig'+str(nfig)+'.'+figext
                   
    #fig=mlab.figure(nfig)
    #old_bg=fig.scene.background
    #fig.scene.background=(1, 1, 1)
    #if verbose:
      #print('  Save Mayavi Scene %d in %s'%(nfig,File))
    #mlab.savefig(File,magnification=scale) 
    #fig.scene.background=old_bg